/******************************************************************
 * PalmOS includes.
 ******************************************************************/
#include <Common.h>
#include <System/SysAll.h>
#include <System/SysEvtMgr.h>
#include <UI/UIAll.h>

/******************************************************************
 * Application includes.
 ******************************************************************/
#include "callback.h"

#include "Timesheet.h"
#include "Globals.h"
#include "Category.h"

/******************************************************************
 * Application category functions.
 ******************************************************************/

/******************************************************************
 * Attempt to delete an existing entry string from the category at
 * the specified point. All transTable entries that correspond to
 * the entry that is deleted will be remapped to 255. Caller is
 * expected to deal with transTable fixup after calling function.
 * Returns the new handle to the category after it has been resized
 * by deleting category.
 ******************************************************************/
VoidHand
CatDeleteEntry(UInt catRecIdx, Byte numEntries, UInt catDelIdx)
{
  VoidHand catHand;
  TSCatRecType *catPtr;
  Byte delMarker;
  Byte trans;
  Byte idx;

  catHand = DmGetRecord(TSDatabase, catRecIdx);
  catPtr = MemHandleLock(catHand);
  /* shuffle the remaining entries down over the deleted entry */
  for (idx = catDelIdx; idx < numEntries - 1; idx += 1)
    {
      DmStrCopy(catPtr, (VoidPtr) (catPtr->cats[idx]) - (VoidPtr) (catPtr), catPtr->cats[idx + 1]);
    }
  /* find the translation table entries for the category we've just deleted */
  delMarker = 255;
  for (idx = 1; idx < TSMaxCatEntries; idx += 1)
    {
      if (catPtr->transTable[idx] == catDelIdx)
        {
          /* mark the deleted translation table index as deleted and let the caller deal with it.
           * note: we don't break out of the loop because multiple transTable entries may point
           * to a single actual entry (which may be the one we've just deleted). */
          DmWrite(catPtr, sizeof(Byte) * idx, &delMarker, sizeof(Byte));
        }
    }
  /* fix up the translation table entries after the deletion point (by subtracting 1!) */
  for (idx = 1; idx < TSMaxCatEntries; idx += 1)
    {
      trans = catPtr->transTable[idx];
      if ((trans >= catDelIdx) && (trans < TSMaxCatEntries))
        {
          trans -= 1;
          DmWrite(catPtr, (VoidPtr) (&(catPtr->transTable[idx])) - (VoidPtr) (catPtr), &trans, sizeof(Byte));
        }
    }
  /* make the category record one entry smaller */
  MemHandleUnlock(catHand);
  DmReleaseRecord(TSDatabase, catRecIdx, true);
  return DmResizeRecord(TSDatabase, catRecIdx, (sizeof(Byte) * TSMaxCatEntries)
                    + (sizeof(Char) * TSMaxCatEntryLen * (numEntries - 1)));
}

/******************************************************************
 * Attempt to insert a new entry string into the category at the
 * appropriate alphabetic point. Returns the index_entry was
 * inserted into, or -1 - index_entry if the entry already exists.
 ******************************************************************/
Int
CatInsertEntry(UInt catRecIdx, Byte numEntries, CharPtr newCat)
{
  UInt idx;
  Int result;
  Byte trans;
  Byte newCatIdx;
  VoidHand catHand;
  TSCatRecType *catPtr;

  /* check there isn't any existing category of the same name, while also searching for the
   * appropriate insertion point for the new entry */
  catHand = DmQueryRecord(TSDatabase, catRecIdx);
  catPtr = MemHandleLock(catHand);
  for (idx = 0, newCatIdx = TSMaxCatEntries; idx < numEntries; idx += 1)
    {
      result = StrCompare(newCat, catPtr->cats[idx]);
      if (result == 0)
        {
          MemHandleUnlock(catHand);
          return -1 - idx;
        }
      else if ((idx > 0) && (idx < numEntries - 1) && (newCatIdx == TSMaxCatEntries) && (result < 0))
        {
          /* have found the appropriate (alphabetic) insertion point for the new category */
          newCatIdx = idx;
          /* note we keep searching for duplicate entries, as last 'Edit...' entry is *not* in alphabetical order */
        }
    }
  /* check we have found insertion point, otherwise default to entry before 'Edit...' */
  if (newCatIdx == TSMaxCatEntries)
    {
      newCatIdx = numEntries - 1;
    }
  MemHandleUnlock(catHand);
  /* make room for insertion of new category string */
  catHand = DmResizeRecord(TSDatabase, catRecIdx, (sizeof(Byte) * TSMaxCatEntries)
                    + (sizeof(Char) * TSMaxCatEntryLen * (numEntries + 1)));
  catPtr = MemHandleLock(catHand);
  for (idx = numEntries; idx > newCatIdx; idx -= 1)
    {
      DmStrCopy(catPtr, (VoidPtr) (catPtr->cats[idx]) - (VoidPtr) (catPtr), catPtr->cats[idx - 1]);
    }
  /* save the field text in the free category */
  DmWrite(catPtr, (VoidPtr) (&(catPtr->cats[newCatIdx])) - (VoidPtr) (catPtr), newCat, sizeof(Byte) * TSMaxCatEntryLen);
  /* fix up the translation table entries after the insertion point (by adding 1!) */
  for (idx = 1; idx < TSMaxCatEntries; idx += 1)
    {
      trans = catPtr->transTable[idx];
      if ((trans >= newCatIdx) && (trans < TSMaxCatEntries))
        {
          trans += 1;
          DmWrite(catPtr, (VoidPtr) (&(catPtr->transTable[idx])) - (VoidPtr) (catPtr), &trans, sizeof(Byte));
        }
    }
  MemHandleUnlock(catHand);
  DmReleaseRecord(TSDatabase, catRecIdx, true);
  return newCatIdx;
}

/******************************************************************
 * Utility function that returns the index of the entry in the
 * transTable that contains the actual idx passed to the function.
 * Used to produce client/project/task transtable indexes from the
 * actual user GUI list selections.
 ******************************************************************/
Word
CatReverseLookup(Word idx, BytePtr transTable, UInt numEntries)
{
  UInt tblIdx = 1;

  /* quick check to see if we should default to unspecified or 'none' entry */
  if (idx <= 0 || idx >= numEntries)
    {
      /* 'none' entry will always be at index 0 of the translation table */
      return 0;
    }
  /* search the translation table for the table index containing the supplied index! */
  for (; tblIdx < numEntries && transTable[tblIdx] != idx; tblIdx += 1)
    {
      /* this body left intentionally empty */
    }
  return tblIdx;
}

/******************************************************************
 * Build an array of pointers to the strings in the specified category
 * (in translation table order). Typically called before
 * the pointer array is used to initialise a GUI list control.
 * Caller is responsible for freeing returned dynamically allocated pointer.
 ******************************************************************/
VoidPtr
CatBuildPtrArray(TSCatRecType * catRec, Byte numEntries)
{
  CharPtr *ptr;
  UInt idx;

  /* dynamically allocate space for the pointer array */
  ptr = MemPtrNew(sizeof(CharPtr) * numEntries);
  if (!ptr)
    {
      /* low on memory! should check or halt program! */
      return NULL;
    }
  /* build a pointer array to pass to LstSetListChoices (is there an LstAddChoice function?) */
  for (idx = 0; idx < numEntries; idx += 1)
    {
      ptr[idx] = catRec->cats[idx];
    }
  return ptr;
}

/******************************************************************
 * Utility function, set up a resource list record for the supplied
 * list. There's three lists that are important to Timesheet, the
 * Client list, the Project list and the Task list.
 ******************************************************************/
Err
InitCatResource(UIntPtr resIdP, Int initStringID, UInt numEntries)
{
  Byte idx;
  VoidHand hand;
  VoidHand resHand;
  Ptr catPtr;

  /* create and initialise a new list record to hold the supplied client, project or task list.
   * Note we only allocate enough space to store the initial default entries */
  hand = DmNewRecord(TSDatabase, resIdP, sizeof(Byte) * TSMaxCatEntries
                     + (sizeof(Char) * TSMaxCatEntryLen * numEntries));
  if (!hand)
    {
      /* failed to create record */
      return appErrorClass;
    }
  catPtr = MemHandleLock(hand);
  /* initialise the translation table in the record */
  DmSet(catPtr, 0, sizeof(Byte) * TSMaxCatEntries, TSMaxCatEntries);
  for (idx = 0; idx < numEntries - 1; idx++)
    {
      DmWrite(catPtr, sizeof(Byte) * idx, &idx, sizeof(Byte));
    }
  /* last transTable entry *always* points to 'Edit...' category name */
  DmWrite(catPtr, sizeof(Byte) * (TSMaxCatEntries - 1), &idx, sizeof(Byte));
  /* write the default category names */
  resHand = DmGet1Resource('tSTR', initStringID);
  DmWrite(catPtr, sizeof(Byte) * TSMaxCatEntries, MemHandleLock(resHand),
          sizeof(Char) * TSMaxCatEntryLen * numEntries);
  MemHandleUnlock(resHand);
  DmReleaseResource(resHand);
  /* have created another resource record */
  MemHandleUnlock(hand);
  DmReleaseRecord(TSDatabase, *resIdP, true);
  (*resIdP) += 1;
  return 0;
}

/*** End of file ***/