/******************************************************************
 * PalmOS includes.
 ******************************************************************/
#include <Common.h>
#include <System/SysAll.h>
#include <System/SysEvtMgr.h>
#include <UI/UIAll.h>

/******************************************************************
 * Application includes.
 ******************************************************************/
#include "callback.h"
#include "resource.h"

#include "Timesheet.h"
#include "Globals.h"
#include "Category.h"
#include "Database.h"

/******************************************************************
 * Application database related functions.
 ******************************************************************/

/******************************************************************
 *
 ******************************************************************/
Int
_SummaryFindCompareFunc(VoidPtr rec1Ptr, VoidPtr rec2Ptr, Int other_ignored)
{
  TSSummaryRecType *summaryRec1Ptr;
  TSSummaryRecType *summaryRec2Ptr;
  Int retVal;

  /* Start fixup from the GNU Palmpilot SDK FAQ so callback function can access global variables */
  CALLBACK_PROLOGUE;

  /* Translate the generic rec pointers into a more useful form. */
  summaryRec1Ptr = rec1Ptr;
  summaryRec2Ptr = rec2Ptr;
  /* Order records by Client / Project / Task indexes */
  if (summaryRec1Ptr->catIdx[0] == summaryRec2Ptr->catIdx[0])
    {
      if (summaryRec1Ptr->catIdx[1] == summaryRec2Ptr->catIdx[1])
        {
          if (summaryRec1Ptr->catIdx[2] == summaryRec2Ptr->catIdx[2])
            {
              /* Records are equal */
              retVal = 0;
            }
          else
            {
              /* Order records by Task */
              retVal = summaryRec1Ptr->catIdx[2] - summaryRec2Ptr->catIdx[2];
            }
        }
      else
        {
          /* Order records by Project */
          retVal = summaryRec1Ptr->catIdx[1] - summaryRec2Ptr->catIdx[1];
        }
    }
  else
    {
      /* Order records by Client */
      retVal = summaryRec1Ptr->catIdx[0] - summaryRec2Ptr->catIdx[0];
    }
  /* End fixup from the GNU Palmpilot SDK FAQ so callback function can access global variables */
  CALLBACK_EPILOGUE;

  return retVal;
}

/******************************************************************
 * Build a temporary summary database. Used to provide weekly and monthly summary views.
 ******************************************************************/
Boolean
BuildSummaryDatabase(DateType * startDate, DateType * endDate)
{
  Err error;
  LocalID summaryID;
  Int summaryCardNum;
  DmOpenRef summaryDB;
  UInt startDayRecIdx;
  UInt endDayRecIdx;
  UInt dayRecIdx;
  UInt summaryRecIdx;
  Boolean summaryRecExists;
  VoidHand entryHand;
  VoidHand summaryHand;
  VoidHand catHand;
  TSEntryRecType *entryPtr;
  TSSummaryRecType *summaryPtr;
  TSCatRecType *catPtr;
  TSSummaryRecType dummySummaryRec;
  TSSummaryTotalRecType summaryTotalRec;
  UInt recAttr;
  UInt numSummaryRecs;
  UInt sumHours;
  UInt sumMins;
  UInt catNum;

  /* Check if there's an existing summary database */
  summaryDB = DmOpenDatabaseByTypeCreator(TSSummaryDBType, TSAppID, dmModeReadWrite);
  if (summaryDB)
    {
      /* Get locaID, cardID etc. so we can delete database */
      DmOpenDatabaseInfo(summaryDB, &summaryID, NULL, NULL, &summaryCardNum, NULL);
      /* Close and delete existing summary database */
      DmCloseDatabase(summaryDB);
      DmDeleteDatabase(summaryCardNum, summaryID);
    }
  /* Create and open new summary database */
  error = DmCreateDatabase(0, TSSummaryDBName, TSAppID, TSSummaryDBType, false);
  if (error)
    {
      /* Failed to create database */
      if (error == dmErrMemError || error == memErrNotEnoughSpace)
        {
          FrmAlert(alertID_tsLowMem);
        }
      return false;
    }
  summaryDB = DmOpenDatabaseByTypeCreator(TSSummaryDBType, TSAppID, dmModeReadWrite);
  numSummaryRecs = 0;
  summaryTotalRec.totalHours = 0;
  summaryTotalRec.totalMins = 0;
  summaryTotalRec.totalChargeableHours = 0;
  summaryTotalRec.totalChargeableMins = 0;
  if (summaryDB)
    {
      /* Convert start/end dates into record indexes */
      if (!DateRangeToDatabaseIndex(startDate, endDate, &startDayRecIdx, &endDayRecIdx))
      {
          /* Nothing to summarise, so we've finished. */
          goto SUMMARY_FINISHED;
        }
      /* Build summary database */
      for (dayRecIdx = startDayRecIdx; dayRecIdx <= endDayRecIdx; dayRecIdx++)
        {
          /* Day or entry record? */
          DmRecordInfo(TSDatabase, dayRecIdx, &recAttr, NULL, NULL);
          if ((recAttr & TSRecAttrMask) == TSEntryRecAttr)
            {
              /* Have another entry record to include in summary database. */
              entryHand = DmQueryRecord(TSDatabase, dayRecIdx);
              entryPtr = MemHandleLock(entryHand);
              /* Build a dummy record to insert */
              MemMove(&dummySummaryRec, entryPtr, sizeof(Byte) * 3);    /* copy client/project/task */
              /* Translate the record's category indexes into actual indexes so summary is sorted alphabetically */
              for (catNum = 0; catNum < 3; catNum++)
                {
                  /* Get the appropriate category record */
                  catHand = DmQueryRecord(TSDatabase, catNum + 1);
                  catPtr = MemHandleLock(catHand);
                  /* Translate the index into an actual category entry */
                  dummySummaryRec.catIdx[catNum] = catPtr->transTable[dummySummaryRec.catIdx[catNum]];
                  MemHandleUnlock(catHand);
                }
              /* Translate entry Byte duration format to summary UInt duration format */
              dummySummaryRec.totalDuration = entryPtr->hours & 0x0F;   /* Hours in low 12 bits */
              dummySummaryRec.totalDuration |= (entryPtr->hours & 0xF0) << 8;   /* Minutes in high 4 bits */
              /* Add the existing record duration to the total duration */
              summaryTotalRec.totalHours += (dummySummaryRec.totalDuration & 0x0FFF);
              summaryTotalRec.totalMins += ((dummySummaryRec.totalDuration & 0xF000) >> 12);
              /* Roll up mins to hours */
              if (summaryTotalRec.totalMins >= 12)
                {
                  summaryTotalRec.totalMins -= 12;
                  summaryTotalRec.totalHours++;
                }
              if (recAttr & TSEntryChargedAttr)
                {
                  /* Add the existing record duration to the total chargeable duration */
                  summaryTotalRec.totalChargeableHours += (dummySummaryRec.totalDuration & 0x0FFF);
                  summaryTotalRec.totalChargeableMins += ((dummySummaryRec.totalDuration & 0xF000) >> 12);
                  /* Roll up mins to hours */
                  if (summaryTotalRec.totalChargeableMins >= 12)
                    {
                      summaryTotalRec.totalChargeableMins -= 12;
                      summaryTotalRec.totalChargeableHours++;
                    }
                }
              MemHandleUnlock(entryHand);
              /* Find the existing summary record, or the insert point for new summary record. */
              summaryRecIdx = DmFindSortPositionV10(summaryDB, &dummySummaryRec,
                               (DmComparF *) (&_SummaryFindCompareFunc), 0);
              /* Check if the summary record already exists or not */
              if (summaryRecIdx > 0 && (summaryRecIdx - 1) < numSummaryRecs)
                {
                  summaryHand = DmQueryRecord(summaryDB, summaryRecIdx - 1);
                  summaryPtr = MemHandleLock(summaryHand);
                  /* summaryRecExists set to true if the record already exists. */
                  if (summaryPtr->catIdx[0] == dummySummaryRec.catIdx[0]
                      && summaryPtr->catIdx[1] == dummySummaryRec.catIdx[1]
                      && summaryPtr->catIdx[2] == dummySummaryRec.catIdx[2])
                    {
                      /* Existing summary record just BEFORE returned summary record index */
                      summaryRecIdx--;
                      summaryRecExists = true;
                    }
                  else
                    {
                      summaryRecExists = false;
                    }
                  MemHandleUnlock(summaryHand);
                }
              else
                {
                  /* Need to append new summary record to end of summary database. */
                  summaryRecExists = false;
                }
              /* Update existing record or create new record? */
              if (summaryRecExists)
                {
                  /* Update existing record. */
                  summaryHand = DmGetRecord(summaryDB, summaryRecIdx);
                  summaryPtr = MemHandleLock(summaryHand);
                  /* Add the existing record duration to the dummy record duration */
                  sumHours = (summaryPtr->totalDuration & 0x0FFF)
                    + (dummySummaryRec.totalDuration & 0x0FFF);
                  sumMins = ((summaryPtr->totalDuration & 0xF000) >> 12)
                    + ((dummySummaryRec.totalDuration & 0xF000) >> 12);
                  /* Roll up mins to hours */
                  if (sumMins >= 12)
                    {
                      sumMins -= 12;
                      sumHours++;
                    }
                  /* Store total duration in dummy record */
                  dummySummaryRec.totalDuration = (sumMins << 12);
                  dummySummaryRec.totalDuration |= sumHours;
                  /* Copy in the updated dummy record */
                  DmWrite(summaryPtr, 0, &dummySummaryRec, sizeof(TSSummaryRecType));
                }
              else
                {
                  /* Create new record. */
                  summaryHand = DmNewRecord(summaryDB, &summaryRecIdx, sizeof(TSSummaryRecType));
                  if (!summaryHand)
                    {
                      /* Get locaID, cardID etc. so we can delete database as we appear to be low on memory */
                      DmOpenDatabaseInfo(summaryDB, &summaryID, NULL, NULL, &summaryCardNum, NULL);
                      /* Close and delete existing summary database */
                      DmCloseDatabase(summaryDB);
                      DmDeleteDatabase(summaryCardNum, summaryID);
                      /* Alert the user the operation failed */
                      FrmAlert(alertID_tsLowMem);
                      return false;
                    }
                  summaryPtr = MemHandleLock(summaryHand);
                  /* Copy in the dummy record */
                  DmWrite(summaryPtr, 0, &dummySummaryRec, sizeof(TSSummaryRecType));
                  numSummaryRecs++;
                }
              MemHandleUnlock(summaryHand);
              DmReleaseRecord(summaryDB, summaryRecIdx, true);
            }
        }
      /* Completed summary database */
    SUMMARY_FINISHED:
      /* Add total hours/minutes record to end of database */
      summaryHand = DmNewRecord(summaryDB, &numSummaryRecs, sizeof(TSSummaryTotalRecType));
      if (!summaryHand)
        {
          /* Get locaID, cardID etc. so we can delete database as we appear to be low on memory */
          DmOpenDatabaseInfo(summaryDB, &summaryID, NULL, NULL, &summaryCardNum, NULL);
          /* Close and delete existing summary database */
          DmCloseDatabase(summaryDB);
          DmDeleteDatabase(summaryCardNum, summaryID);
          /* Alert the user the operation failed */
          FrmAlert(alertID_tsLowMem);
          return false;
        }
      /* Save the summary database total hours and minutes */
      summaryPtr = MemHandleLock(summaryHand);  /* Just reusing pointer */
      DmWrite(summaryPtr, 0, &summaryTotalRec, sizeof(TSSummaryTotalRecType));
      MemHandleUnlock(summaryHand);
      DmCloseDatabase(summaryDB);
      return true;
    }
  else
    {
      /* Failed to open created summary database for some reason (we'll assume it's low memory) */
      FrmAlert(alertID_tsLowMem);
      return false;
    }
}

/******************************************************************
 * Return the number of summary database records (if any).
 * Note the returned count includes the last 'totals' record in an
 * existing summary database.
 ******************************************************************/
Int GetNumSummaryRecords(void)
{
  DmOpenRef summaryDB;
  Int numSummaryRecs;

  /* Check if there's an existing summary database */
  summaryDB = DmOpenDatabaseByTypeCreator(TSSummaryDBType, TSAppID, dmModeReadWrite);
  if (summaryDB)
    {
      /* Get the number of records in the database */
      numSummaryRecs = DmNumRecords(summaryDB);
      DmCloseDatabase(summaryDB);
    }
  else
    {
      /* No records to display */
      numSummaryRecs = 0;
    }
  return numSummaryRecs;
}

/******************************************************************
 * Convert a start and end date into indexes in the Timesheet database.
 * Returns false if no actual records fall within range.
 * Returns true if at least one day record falls within range.
 ******************************************************************/
Boolean DateRangeToDatabaseIndex(DateType * startDate, DateType * endDate,
                                        UInt * startRecIdx, UInt * endRecIdx)
{
  Boolean startRecExists;
  Boolean endRecExists;
  VoidHand dayHand;
  TSDayRecType *dayPtr;

  /* Find start day index */
  FindDayRec(startDate, startRecIdx, &startRecExists);
  /* Find end day index */
  FindDayRec(endDate, endRecIdx, &endRecExists);
  /* Check if we have anything to do */
  if ((*startRecIdx) == (*endRecIdx) && !(startRecExists || endRecExists))
    {
      /* No records fall between start and end date */
      return false;
    }
  /* Find the adjusted endRecIdx */
  if (endRecExists)
    {
      /* Find number of entries in existing end day */
      dayHand = DmQueryRecord(TSDatabase, (*endRecIdx));
      dayPtr = MemHandleLock(dayHand);
      (*endRecIdx) += dayPtr->numEntries;
      MemHandleUnlock(dayHand);
    }
  else
    {
      /* End day doesn't exist, go to end of previous day. */
      (*endRecIdx)--;
    }
  /* Have at least one day of records in range */
  return true;
}

/******************************************************************
 * Erase all database records that fall between start and end date inclusive
 ******************************************************************/
void
EraseDatabaseRange(DateType * startDate, DateType * endDate)
{
  UInt startRecIdx;
  UInt endRecIdx;
  UInt recIdx;

  /* Convert start and end dates into database record indexes */
  if (DateRangeToDatabaseIndex(startDate, endDate, &startRecIdx, &endRecIdx))
    {
      /* Have some records to delete */
      for(recIdx = startRecIdx; recIdx <= endRecIdx; recIdx ++)
        {
          /* That's not a typo, we just keep deleting the start record and following records
           * get shuffled up. */
          DmRemoveRecord(TSDatabase, startRecIdx);
        }
      /* If we're timing and entry, check if we just deleted it */
      if (TSEntryTimer.recIdx > 0)
      {
        if (TSEntryTimer.recIdx >= startRecIdx)
        {
          if (TSEntryTimer.recIdx <= endRecIdx)
          {
            /* Timed entry has been deleted, clear timer */
            TSEntryTimer.recIdx = 0;
            TSEntryTimer.secs = 0;
          }
          else
          {
            /* Timed entry has not been deleted, but does need to be adjusted down as
             * intermediate records have been deleted */
            TSEntryTimer.recIdx -= (endRecIdx - startRecIdx) + 1;
          }
        }
        /* else do nothing, timed record appears BEFORE start deletion index */
      }
    }
}

/******************************************************************
 * Record compare function used by the DmFindSortPosition function.
 ******************************************************************/
Int
_DayFindCompareFunc(VoidPtr rec1Ptr, VoidPtr rec2Ptr, Int other_ignored,
                    SortRecordInfoPtr rec1InfoPtr, SortRecordInfoPtr rec2InfoPtr, VoidHand appInfoH)
{
  Int retVal;
  Byte rec1Type;
  Byte rec2Type;
  VoidHand rec2Hand;
  Long UID;
  UInt rec2ID;

  /* Start fixup from the GNU Palmpilot SDK FAQ so callback function can access global variables */
  CALLBACK_PROLOGUE;

  rec2ID = 0;
  /* find the Timesheet record types of the two records we've been passed. */
  rec1Type = rec1InfoPtr->attributes & TSRecAttrMask;
  rec2Type = rec2InfoPtr->attributes & TSRecAttrMask;
  /* act depending on the record's first two attribute bits. These bits uniquely identify all record types
   * presently used in the Timesheet database */
  if (rec2Type == TSSpecialRecAttr)
    {
      /* special record in the database (always come 'first') */
      retVal = 1;
    }
  else
    {
      if (rec2Type == TSEntryRecAttr)
        {
          /* day entry records in the database, must find the earlier DAY header record this entry corresponds to */
          /* find the ID this record corresponds to */
          UID = (rec2InfoPtr->uniqueID[2])
            | ((Long) rec2InfoPtr->uniqueID[1] << 8)
            | ((Long) rec2InfoPtr->uniqueID[0] << 16);
          DmFindRecordByID(TSDatabase, UID, &rec2ID);
          /* search back to get the day this entry lives under */
          rec2ID -= ((TSEntryRecType *) rec2Ptr)->entryNum;
          /* load and lock the day record */
          rec2Hand = DmQueryRecord(TSDatabase, rec2ID);
          rec2Ptr = MemHandleLock(rec2Hand);
        }
      /* have two day headers to compare, so compare them and set the return result */
      retVal = DateToDays(((TSDayRecType *) rec1Ptr)->dayDate) - DateToDays(((TSDayRecType *) rec2Ptr)->dayDate);
      /* release any records we had to search for */
      if (rec2ID != 0)
        {
          MemHandleUnlock(rec2Hand);
        }
    }

  /* end fixup from the GNU Palmpilot SDK FAQ so callback function can access global variables */
  CALLBACK_EPILOGUE;

  return retVal;
}

/******************************************************************
 *
 ******************************************************************/
Boolean
FindEarliestDayRec(void)
{
  VoidHand recHand;
  TSDayRecType *recPtr;

  /* Check the database has at least one day in it */
  if (DmNumRecords(TSDatabase) > TSFirstDayRecIdx)
    {
      /* Change the current date to the earliest (first) day in the database */
      recHand = DmQueryRecord(TSDatabase, TSFirstDayRecIdx);
      recPtr = MemHandleLock(recHand);
      TSPresentDayDate.day = recPtr->dayDate.day;
      TSPresentDayDate.month = recPtr->dayDate.month;
      TSPresentDayDate.year = recPtr->dayDate.year;
      MemHandleUnlock(recHand);
      /* Found earliest day */
      return true;
    }
  else
    {
      /* No earliest day */
      return false;
    }
}

/******************************************************************
 *
 ******************************************************************/
Boolean
FindLatestDayRec(void)
{
  VoidHand recHand;
  TSDayRecType *recPtr;

  /* Check the database has at least one day in it */
  if (DmNumRecords(TSDatabase) > TSFirstDayRecIdx)
    {
      /* Change the current date to the latest (last) day in the database */
      recHand = DmQueryRecord(TSDatabase, FindPrevDayRecIdx(TSDatabase, DmNumRecords(TSDatabase) - 1));
      recPtr = MemHandleLock(recHand);
      TSPresentDayDate.day = recPtr->dayDate.day;
      TSPresentDayDate.month = recPtr->dayDate.month;
      TSPresentDayDate.year = recPtr->dayDate.year;
      MemHandleUnlock(recHand);
      return true;
    }
  else
    {
      /* No later day */
      return false;
    }
}

/******************************************************************
 * Finds TSPresentDayRecIdx for the day corresponds to the date
 * held in TSPresentDayDate using the Data Manager binary search.
 * Also sets up TSPresentDayRecIdx to the appropriate value.
 ******************************************************************/
void
FindDayRec(DateType * dayDate, UInt * presentDayRecIdx, Boolean * presentDayRecExists)
{
  TSDayRecType dummyDayRec;
  SortRecordInfoType dummyDayInfo;
  VoidHand dayRecHand;
  TSDayRecType *dayRecPtr;
  UInt searchRecIdx;

  /* Check if we have any records to search */
  if (DmNumRecords(TSDatabase) == TSFirstDayRecIdx)
    {
      /* Nope, don't bother searching */
      *presentDayRecIdx = TSFirstDayRecIdx;
      *presentDayRecExists = false;
      return;
    }
  /* set up the 'new' record for searching */
  MemMove(&(dummyDayRec.dayDate), dayDate, sizeof(DateType));
  dummyDayInfo.attributes = TSDayRecAttr;
  /* Use Data Manager's binary search. Requires a moderately intelligent comparison function due to database structure. */
  searchRecIdx = DmFindSortPosition(TSDatabase, &dummyDayRec, &dummyDayInfo, (DmComparF *) (&_DayFindCompareFunc), 0);
  /* Record may or may not already exist, have to perform one more test to see what search returned. */
  /* Find the day entry the found record belongs in */
  *presentDayRecIdx = FindPrevDayRecIdx(TSDatabase, searchRecIdx - 1);
  /* Load and lock... */
  dayRecHand = DmQueryRecord(TSDatabase, *presentDayRecIdx);
  dayRecPtr = MemHandleLock(dayRecHand);
  /* does it match the search date? */
  if (DateToDays(dayRecPtr->dayDate) != DateToDays(*dayDate))
    {
      /* revert to original returned index */
      *presentDayRecIdx = searchRecIdx;
      *presentDayRecExists = false;
    }
  else
    {
      /* found matching day */
      *presentDayRecExists = true;
    }
  MemHandleUnlock(dayRecHand);
}

/******************************************************************
 * Find the previous day record index from a given record index.
 * Note database to search is specified as argument because this
 * function is used by the FindInDatabase function (which is invoked
 * by the Palmpilot 'Find' button).
 ******************************************************************/
UInt
FindPrevDayRecIdx(DmOpenRef db, UInt presentRecIdx)
{
  VoidHand entryHand;
  TSEntryRecType *entryPtr;
  UInt attr;

  /* sanity check required for binary search */
  if (presentRecIdx < TSFirstDayRecIdx)
    {
      return TSFirstDayRecIdx;
    }
  /* first check if we're starting on a day */
  DmRecordInfo(db, presentRecIdx, &attr, NULL, NULL);
  if ((attr & 0x07) == TSDayRecAttr)
    {
      /* already on a day! */
      return presentRecIdx;
    }
  /* go back to the day record for the current entry */
  entryHand = DmQueryRecord(db, presentRecIdx);
  entryPtr = MemHandleLock(entryHand);
  presentRecIdx -= entryPtr->entryNum;
  MemHandleUnlock(entryHand);
  return presentRecIdx;
}

/******************************************************************
 * Retrieve the current application preferences from the opened Timesheet database.
 ******************************************************************/
Err OpenPreferences(void)
{
  VoidHand hand;
  TSAppPrefType *prefPtr;

  /* Retreive the application preference info from record zero of the database */
  hand = DmQueryRecord(TSDatabase, 0);
  prefPtr = MemHandleLock(hand);
  /* Load global copy of application preferences */
  MemMove(&TSAppPrefs, prefPtr, sizeof(TSAppPrefType));
  /* Check if we have a timer record on the end */
  if (MemHandleSize(hand) >= sizeof(TSAppPrefType) + sizeof(TSEntryTimerType))
  {
    /* Load global timer record as well */
    MemMove(&TSEntryTimer, ((void*)prefPtr) + sizeof(TSAppPrefType), sizeof(TSEntryTimerType));
    MemHandleUnlock(hand);
  }
  else
  {
    /* No timer record, so clear global timer record */
    MemSet(&TSEntryTimer, sizeof(TSEntryTimerType), 0);
    /* Resize the preferences record to include at least one timer structure */
    MemHandleUnlock(hand);
    hand = DmResizeRecord(TSDatabase, 0, sizeof(TSAppPrefType) + sizeof(TSEntryTimerType));
    if (!hand)
    {
      /* Failed to resize record */
      FrmAlert(alertID_tsLowMem);
      return appErrorClass;
    }
    DmReleaseRecord(TSDatabase, 0, false);
  }
  /* Opened preferences ok */
  return 0;
}

/******************************************************************
 * Save the current application preferences back to the open Timesheet database, typically
 * prior to application close.
 ******************************************************************/
void ClosePreferences(void)
{
  VoidHand hand;
  VoidPtr voidPtr;

  /* Save the application preferences back to the timesheet resource database */
  hand = DmGetRecord(TSDatabase, 0);
  voidPtr = MemHandleLock(hand);
  DmWrite(voidPtr, 0, &TSAppPrefs, sizeof(TSAppPrefType));
  /* Sanity check the timer */
  if (TSEntryTimer.recIdx == 0 && TSEntryTimer.secs > 0)
  {
        /* HACK! Timing non-saved entry, stop timer */
        TSEntryTimer.recIdx = 0;
        TSEntryTimer.secs = 0;
  }
  /* If we get this far we can be sure there's room for at least one timer record in the
   * preferences record as OpenPrefences will have expanded the prefs record if required */
  DmWrite(voidPtr, sizeof(TSAppPrefType), &TSEntryTimer, sizeof(TSEntryTimerType));
  MemHandleUnlock(hand);
  DmReleaseRecord(TSDatabase, 0, true);
}

/******************************************************************
 * Used by the FindInDatabase function below to search entry categories for matching
 * client/project/task names.
 ******************************************************************/
Boolean
FindInCategory(DmOpenRef db, UInt catRec, UInt entryCatIdx, FindParamsPtr findParams)
{
  Boolean match;
  VoidHand catHand;
  TSCatRecType *catPtr;
  CharPtr catDescPtr;
  Word pos;

  catHand = DmQueryRecord(db, catRec);
  catPtr = MemHandleLock(catHand);
  catDescPtr = catPtr->cats[catPtr->transTable[entryCatIdx]];
  match = FindStrInStr(catDescPtr, findParams->strToFind, &pos);
  MemHandleUnlock(catHand);
  return match;
}

/******************************************************************
 * Search the Timesheet database for matching records. Called by PilotMain
 * to implement the Find command. Basically lifted from Memopad.c, thanks Palm computing!
 ******************************************************************/
void
FindInDatabase(FindParamsPtr findParams)
{
  Boolean done;
  Boolean fits;
  Int textLen;
  Int width;
  UInt recIdx;
  UInt numRecs;
  UInt recAttr;
  UInt cardNo;
  LocalID dbID;
  Word pos;
  VoidHand recordHand;
  VoidPtr recordPtr;
  Boolean match;
  CharPtr entryDescPtr;
  DmOpenRef tsDatabase;
  DateType dayDate;
  FontID currFont;
  RectangleType rect;
  Char dateStr[20 + 1];
  SystemPreferencesType sysPrefs;

  /* Get a copy of the system's preferences */
  PrefGetPreferences(&sysPrefs);
  /* Mark dayDate as unfilled (yet). Required as we may begin searching from middle of entries for certain day, so
   * won't have day date for any matching entries, see code below. */
  dayDate.day = 0;
  /* Find the app's database. */
  tsDatabase = DmOpenDatabaseByTypeCreator(TSDBType, TSAppID, findParams->dbAccesMode);
  if (!tsDatabase)
    {
      /* Return without anything, also indicate that no more records are expected. */
      findParams->more = false;
      return;
    }
  numRecs = DmNumRecords(tsDatabase);
  DmOpenDatabaseInfo(tsDatabase, &dbID, 0, 0, &cardNo, 0);
  /* Display the heading line */
  recordHand = DmGetResource(strRsc, stringID_tsFindHeader);
  recordPtr = MemHandleLock(recordHand);
  done = FindDrawHeader(findParams, (CharPtr) recordPtr);
  MemHandleUnlock(recordHand);
  if (done)
    {
      /* There was no more room to display the heading line */
      DmCloseDatabase(tsDatabase);
      return;
    }
  /* Search the Timesheet entries for the string, starting with the recIdx passed.
   * This allows the search code to be called multiple times when more than one screen
   * of records are found. */
  recIdx = findParams->recordNum;
  /* Skip the first 4 database records as they contain app preferences and category information. */
  if (recIdx < TSFirstDayRecIdx)
    {
      recIdx = TSFirstDayRecIdx;
      findParams->recordNum = TSFirstDayRecIdx;
    }
  while (true)
    {
      /* Because applications can take a long time to finish a find when
       * the result may be on the screen or for other reasons, users like
       * to be able to stop the find.  Stop the find if an event is pending.
       * This stops if the user does something with the device.  Because
       * this call slows down the search we perform it every so many
       * records instead of every record.  The response time should still
       * be short without introducing much extra work to the search.
       * Every 16 records is guessed at to check often enough to respond
       * quickly to the user without wasting time checking too often. */
      if ((recIdx & 0x000f) == 0 && EvtSysEventAvail(true))
        {
          /* Stop the search process. */
          findParams->more = true;
          break;
        }
      /* Have we run out of records? */
      if (recIdx >= numRecs)
        {
          findParams->more = false;
          break;
        }
      /* Get the next record attributes. */
      DmRecordInfo(tsDatabase, recIdx, &recAttr, NULL, NULL);
      recordHand = DmQueryRecord(tsDatabase, recIdx);
      recordPtr = MemHandleLock(recordHand);
      if ((recAttr & 0x07) == TSDayRecAttr)
        {
          /* On day record, don't do any searching, but save the date of the latest day reached. */
          dayDate.day = ((TSDayRecType *) recordPtr)->dayDate.day;
          dayDate.month = ((TSDayRecType *) recordPtr)->dayDate.month;
          dayDate.year = ((TSDayRecType *) recordPtr)->dayDate.year;
          MemHandleUnlock(recordHand);
        }
      else
        {
          /* On entry record, search the categories and entry description for matching string. */
          /* Check the client entry */
          match = FindInCategory(tsDatabase, 1, ((TSEntryRecType *) recordPtr)->clientIdx, findParams);
          /* Check the project entry */
          match |= FindInCategory(tsDatabase, 2, ((TSEntryRecType *) recordPtr)->projectIdx, findParams);
          /* Check the task entry */
          match |= FindInCategory(tsDatabase, 3, ((TSEntryRecType *) recordPtr)->taskIdx, findParams);
          /* Search for the string passed, if it's found display the day date of the entry.
           * An application should call FindStrInStr for each part of the record which
           * needs to be searched. */
          entryDescPtr = (CharPtr) (recordPtr + sizeof(TSEntryRecType));
          match |= FindStrInStr(entryDescPtr, findParams->strToFind, &pos);
          if (match)
            {
              /* Add the match to the find parameter block,  if there is no room to
               * display the match the following function will return true. */
              done = FindSaveMatch(findParams, recIdx, pos, 0, 0, cardNo, dbID);
              if (!done)
                {
                  /* Do we have a day date to display? */
                  if (dayDate.day == 0)
                    {
                      /* No, so find the date of the day record this entry corresponds to. */
                      /* Unload entry record first as FindPrevDayRecIdx queries it. */
                      MemHandleUnlock(recordHand);
                      recordHand = DmQueryRecord(tsDatabase, FindPrevDayRecIdx(tsDatabase, recIdx));
                      recordPtr = MemHandleLock(recordHand);
                      /* Save date from day record */
                      dayDate.day = ((TSDayRecType *) recordPtr)->dayDate.day;
                      dayDate.month = ((TSDayRecType *) recordPtr)->dayDate.month;
                      dayDate.year = ((TSDayRecType *) recordPtr)->dayDate.year;
                      MemHandleUnlock(recordHand);
                      /* Relock entry record and continue */
                      recordHand = DmQueryRecord(tsDatabase, recIdx);
                      recordPtr = MemHandleLock(recordHand);
                      entryDescPtr = (CharPtr) (recordPtr + sizeof(TSEntryRecType));
                    }
                  /* Get the bounds of the region where we will draw the results. */
                  FindGetLineBounds(findParams, &rect);
                  /* Display the record in the search dialog.  We move in one pixel
                   * so that when the record is inverted the left edge is solid. */
                  rect.topLeft.x++;
                  rect.extent.x--;
                  /* Set the standard font, save the current font. */
                  currFont = FntSetFont(stdFont);

                  /* Draw the date string first */
                  DateToAscii(dayDate.month, dayDate.day, dayDate.year + 1920, sysPrefs.dateFormat, dateStr);
                  textLen = StrLen(dateStr);
                  width = FntCharsWidth(dateStr, textLen) + 4;
                  WinDrawChars(dateStr, textLen, rect.topLeft.x, rect.topLeft.y);
                  rect.topLeft.x += width;
                  rect.extent.x -= width;

                  /* Draw start of the entry description */
                  width = rect.extent.x - 2;
                  textLen = StrLen(entryDescPtr);
                  FntCharsInWidth(entryDescPtr, &width, &textLen, &fits);
                  /* Now draw the text from the record. */
                  WinDrawChars(entryDescPtr, textLen, rect.topLeft.x, rect.topLeft.y);
                  /* Restore the font. */
                  FntSetFont(currFont);
                  /* The line number needs to be increment since a line is used for the record. */
                  findParams->lineNumber++;
                }
            }
          MemHandleUnlock(recordHand);
          if (done)
            {
              break;
            }
        }
      recIdx += 1;
    }
  /* Finished searching the database */
  DmCloseDatabase(tsDatabase);
}

/******************************************************************
 * Initialise a new Timesheet database with a small set of
 * application resource records.
 ******************************************************************/
Err
InitDatabase(void)
{
  Err error;
  VoidHand hnd;
  Ptr ptr;
  UInt resId = 0;
  TSAppPrefType defaultPref =
  {
    {TSNumDefaultClients, TSNumDefaultProjects, TSNumDefaultTasks},
    {0, 0, 0}, TSPrefDefaultFlags, 0
    };

  /* create a new record 0 in the database (used to hold the application info) */
  hnd = DmNewRecord(TSDatabase, &resId, sizeof(TSAppPrefType));
  if (!hnd)
    {
      /* failed to allocate record */
      return appErrorClass;
    }
  ptr = MemHandleLock(hnd);
  DmWrite(ptr, 0, &defaultPref, sizeof(TSAppPrefType));
  MemHandleUnlock(hnd);
  DmReleaseRecord(TSDatabase, resId, true);
  /* create a new record in the database to hold the client list */
  resId += 1;
  error = InitCatResource(&resId, stringID_tsDefaultClients, TSNumDefaultClients);
  if (error)
    {
      /* failed to init record */
      return error;
    }
  /* create a new record in the database to hold the project list */
  error = InitCatResource(&resId, stringID_tsDefaultProjects, TSNumDefaultProjects);
  if (error)
    {
      /* failed to init record */
      return error;
    }
  /* create a new record in the database to hold the task list */
  error = InitCatResource(&resId, stringID_tsDefaultTasks, TSNumDefaultTasks);
  if (error)
    {
      /* failed to init record */
      return error;
    }
  /* resource database initialised */
  return 0;
}

/******************************************************************
 * Open and existing application database, or create a new one (with sensible defaults) if none is found.
 ******************************************************************/
Err
OpenDatabase(void)
{
  Err error = 0;
  LocalID dbID;
  UInt dbAttr;
  UInt dbVer;
  UInt cardNo;

  /* open or create the timesheet database for the day and entry records */
  TSDatabase = DmOpenDatabaseByTypeCreator(TSDBType, TSAppID, dmModeReadWrite);
  if (!TSDatabase)
    {
      /* failed to open existing database, so create an new database */
      error = DmCreateDatabase(0, TSDBName, TSAppID, TSDBType, false);
      if (error)
        {
          /* failed to create database */
          return error;
        }
      /* open the new database */
      TSDatabase = DmOpenDatabaseByTypeCreator(TSDBType, TSAppID, dmModeReadWrite);
      if (!TSDatabase)
        {
          /* failed to open database */
          return appErrorClass;
        }
      /* set the database attributes (version, creation date and backup flag) */
      DmOpenDatabaseInfo(TSDatabase, &dbID, NULL, NULL, &cardNo, NULL);
      dbAttr = dmHdrAttrBackup;
      dbVer = TSDBVersion;
      DmSetDatabaseInfo(cardNo, dbID, NULL, &dbAttr, &dbVer, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
      /* initialise the new database with resource records */
      error = InitDatabase();
      if (error)
        {
          /* failed to initialise opened database */
          DmCloseDatabase(TSDatabase);
          DmDeleteDatabase(cardNo, dbID);
          return error;
        }
    }
  else
    {
      /* version check the existing Timesheet database */
      if (VersionCheckDatabase() != 0)
        {
          /* existing database is incompatible */
          DmCloseDatabase(TSDatabase);
          return appErrorClass;
        }
    }
  /* Get the application global preferences from opened database */
  error = OpenPreferences();
  return error;
}

/******************************************************************
 * Compress the category translation table and re-map all existing
 * database entries if required. Return true if some reorging has
 * been performed (generating free category entries), or false if
 * the category is completely full (without any redundant entries).
 ******************************************************************/
Boolean
ReorgDatabase(TSCatRecType * catPtr, Int catRecIdx)
{
  Byte idx;
  Byte catIdx;
  Byte compressCount;
  UInt recIdx;
  UInt numEntries;
  Byte revTransTable[TSMaxCatEntries];
  Byte reorgTransTable[TSMaxCatEntries];
  Byte newTransTable[TSMaxCatEntries];
  VoidHand hand;
  TSDayRecType *dayPtr;
  TSEntryRecType *entryPtr;

  /* init reverse trans table, reorg trans table and new trans table */
  for (idx = 0; idx < TSMaxCatEntries; idx += 1)
    {
      revTransTable[idx] = TSMaxCatEntries;
      newTransTable[idx] = TSMaxCatEntries;
      reorgTransTable[idx] = 0;
    }
  /* build a reverse trans table */
  for (idx = 0, compressCount = 0; idx < TSMaxCatEntries; idx += 1)
    {
      catIdx = catPtr->transTable[idx];
      if (catIdx != TSMaxCatEntries)
        {
          if (revTransTable[catIdx] == TSMaxCatEntries)
            {
              /* nothing entered for this category yet, so fill it */
              revTransTable[catIdx] = idx;
            }
          else
            {
              /* this entry has already been filled, ignore (ie. squeeze out repeat entries) */
              compressCount += 1;
            }
        }                       /* else empty entry? why are we performing reorg? ignore */
    }
  /* have we compressed trans table at all? */
  if (compressCount == 0)
    {
      /* no, don't bother doing any more work, category is completely full. */
      return false;
    }
  /* build a reorg trans table */
  for (idx = 0; idx < TSMaxCatEntries; idx += 1)
    {
      /* mapping from old transTable index to new (compressed) transTable index */
      reorgTransTable[idx] = revTransTable[catPtr->transTable[idx]];
    }
  /* reorg existing transtable */
  for (catIdx = 0; catIdx < TSMaxCatEntries; catIdx += 1)
    {
      if (revTransTable[catIdx] != TSMaxCatEntries)
        {
          newTransTable[revTransTable[catIdx]] = catIdx;
        }
    }
  /* apply reorg trans table to database entries */
  recIdx = TSFirstDayRecIdx;
  while (recIdx < DmNumRecords(TSDatabase))
    {
      /* reorg every entry for current day record */
      hand = DmQueryRecord(TSDatabase, recIdx);
      dayPtr = MemHandleLock(hand);
      numEntries = dayPtr->numEntries;
      MemHandleUnlock(hand);
      for (recIdx += 1; numEntries > 0; numEntries -= 1, recIdx += 1)
        {
          hand = DmGetRecord(TSDatabase, recIdx);
          entryPtr = MemHandleLock(hand);
          catIdx = *(((BytePtr) entryPtr) + (catRecIdx - 1));
          DmWrite(entryPtr, (catRecIdx - 1) * sizeof(Byte), &reorgTransTable[catIdx], sizeof(Byte));
          MemHandleUnlock(hand);
          DmReleaseRecord(TSDatabase, recIdx, true);
        }
    }
  /* reorg the preference saved auto categories */
  TSAppPrefs.newCatEntryIdx[catRecIdx - 1] = reorgTransTable[TSAppPrefs.newCatEntryIdx[catRecIdx - 1]];
  /* save the updated translation table */
  DmWrite(catPtr, 0, newTransTable, sizeof(Byte) * TSMaxCatEntries);
  return true;
}

/******************************************************************
 * Upgrade the existing timesheet database if possible. Returns 0
 * if the database has been upgraded successfully to be compatible
 * with this version of the application, otherwise returns 1 if the
 * database cannot be upgraded.
 *
 * This function is an awful, awful hack...make it a seperate upgrade utility?
 *
 * The Timesheet database has changed as follows:
 * v1.0 -> v1.1: Minor changes, hours/minutes and chargeable flag, all required no database modification.
 * v1.1 -> v1.2: Significant changes in the Client/Project/Task category records. The category pointer array
 *               was removed from the record and is now build dynamically as required. Also, now only the
 *               actual existing entries are stored in the record (whereas blank entries also used to be
 *               stored). This has database category records considerably smaller. Category names were also
 *               changed from 15+1 characters in length to 17+1 characters in length.
 * v1.2 -> v1.3: Expanded app prefs record. THIS VERSION OF TIMESHEET NEVER RELEASED.
 * v1.3 -> v1.4: Added extra 2 bytes to each day entry within each day to support modified
 *               search algorithms.
 * v1.4 -> v1.5: Expanded categories to a maximum of 102 entries per category.
 ******************************************************************/
Err
UpgradeDatabase(UInt cardNo, LocalID dbID, UInt fromDBVersion)
{
  ULong recLen;
  UInt recIdx;
  UInt oldOffset;
  UInt newOffset;
  Byte idx;
  UInt recAttr;
  VoidHand voidHand;
  VoidPtr voidPtr;
  VoidPtr oldCatNamesPtr;
  Byte numEntries[3];

  if (fromDBVersion == 0x0100 /* Timesheet database version 1.0. */  ||
      fromDBVersion == 0x0101 /* Timesheet database version 1.1. */ )
    {
      /* find out how many category entries exist for Client/Project/Task */
      voidHand = DmQueryRecord(TSDatabase, 0);
      voidPtr = MemHandleLock(voidHand);
      MemMove(numEntries, voidPtr, sizeof(Byte) * 3);
      MemHandleUnlock(voidHand);
      /* resize each category record to remove pointer array and empty category names */
      for (recIdx = 1; recIdx <= 3; recIdx++)
        {
          /* take a copy of the old, shorter (15+1) category names. Note there will ALWAYS be at least two
           * category names, 'none' and 'Edit...'. */
          oldOffset = sizeof(Char) * (15 + 1) * numEntries[recIdx - 1];
          oldCatNamesPtr = MemPtrNew(oldOffset);
          if (!oldCatNamesPtr)
            {
              /* appear to be out of memory! */
              return appErrorClass;
            }
          voidHand = DmQueryRecord(TSDatabase, recIdx);
          voidPtr = MemHandleLock(voidHand);
          MemMove(oldCatNamesPtr, voidPtr + (sizeof(Byte) * 22) /* start past 22 entry transTable */ , oldOffset);
          MemHandleUnlock(voidHand);
          /* resize the category record, removing all the excess baggage */
          voidHand = DmResizeRecord(TSDatabase, recIdx, (sizeof(Byte) * TSMaxCatEntries)        /* trans table */
              + (sizeof(Char) * TSMaxCatEntryLen * numEntries[recIdx - 1]));
          voidPtr = MemHandleLock(voidHand);
          /* copy the old categories into their new slots */
          oldOffset = 0;
          newOffset = sizeof(Byte) * TSMaxCatEntries;
          for (idx = 0; idx < numEntries[recIdx - 1]; idx++, oldOffset += (15 + 1), newOffset += TSMaxCatEntryLen)
            {
              DmWrite(voidPtr, newOffset, oldCatNamesPtr + oldOffset, (15 + 1));
            }
          /* updated this category */
          MemPtrFree(oldCatNamesPtr);
          /* if this is the Projects category, need to make SURE 'Edit Projects...' ends with a NULL. This is the
           * whole reason category name length was bumped from 15+1 to 17+1 characters! I.e. Count the characters
           * in 'Edit Projects...'. */
          if (recIdx == 2)
            {
              DmSet(voidPtr, MemHandleSize(voidHand) - 2, 1, 0);
            }
          MemHandleUnlock(voidHand);
          DmReleaseRecord(TSDatabase, recIdx, true);
        }
      /* database has been upgraded to 1.2 format */
      fromDBVersion = 0x0102;
    }
  /* database has either been upgraded to v1.2, or is some version we don't know about */
  if (fromDBVersion == 0x0102 /* Timesheet database version 1.2. */ )
    {
      /* need to expand app prefs record */
      if ((voidHand = DmResizeRecord(TSDatabase, 0, sizeof(TSAppPrefType))) == NULL)
        {
          /* failed to resize prefs record */
          return appErrorClass;
        }
      /* fill rest of resized record with zeros */
      voidPtr = MemHandleLock(voidHand);
      DmSet(voidPtr, sizeof(Byte) * 3, sizeof(TSAppPrefType) - (sizeof(Byte) * 3), 0);
      /* set default pref flags byte */
      DmSet(voidPtr, sizeof(Byte) * 6, sizeof(Byte), TSPrefDefaultFlags);
      MemHandleUnlock(voidHand);
      /* database has been upgraded to 1.3 format */
      fromDBVersion = 0x0103;
    }
  if (fromDBVersion == 0x0103 /* Timesheet database version 1.3. */ )
    {
      /* upgrade database to 1.4 format by expanding each day entry by two bytes */
      for (recIdx = TSFirstDayRecIdx; recIdx < DmNumRecords(TSDatabase); recIdx++)
        {
          DmRecordInfo(TSDatabase, recIdx, &recAttr, NULL, NULL);
          if ((recAttr & TSRecAttrMask) == TSDayRecAttr)
            {
              /* day record, do nothing except resetting idx */
              idx = 1;
            }
          else
            {
              /* entry record, must be expanded to include two new bytes */
              recLen = MemHandleSize(DmQueryRecord(TSDatabase, recIdx));
              voidHand = DmResizeRecord(TSDatabase, recIdx, recLen + (2 * sizeof(Byte)));
              if (!voidHand)
                {
                  /* failed to resize record! */
                  return appErrorClass;
                }
              /* shuffle tail of record down two bytes...hopefully DmWrite handles overlapping ranges */
              voidPtr = MemHandleLock(voidHand);
              DmWrite(voidPtr, 6 * sizeof(Byte), voidPtr + (4 * sizeof(Byte)), recLen - (4 * sizeof(Byte)));
              /* save day entry idx in newly created byte (other byte is a filler) */
              DmSet(voidPtr, 4 * sizeof(Byte), sizeof(Byte), idx);
              MemHandleUnlock(voidHand);
              DmReleaseRecord(TSDatabase, recIdx, true);
              idx += 1;
            }
        }
      /* database has been upgraded to 1.4 format */
      fromDBVersion = 0x0104 /* Timesheet database version 1.4. */ ;
    }
  if (fromDBVersion == 0x0104 /* Timesheet database version 1.4. */ )
    {
      /* Find out how many category entries exist for Client/Project/Task */
      voidHand = DmQueryRecord(TSDatabase, 0);
      voidPtr = MemHandleLock(voidHand);
      MemMove(numEntries, voidPtr, sizeof(Byte) * 3);
      MemHandleUnlock(voidHand);
      /* Expand each category record to add room for 102 entries */
      for (recIdx = 1; recIdx <= 3; recIdx++)
        {
          voidHand = DmResizeRecord(TSDatabase, recIdx,
                                    (sizeof(Byte) * TSMaxCatEntries) +
              ((sizeof(Byte) * TSMaxCatEntryLen) * numEntries[recIdx - 1]));
          if (!voidHand)
            {
              /* Failed to resize record! */
              return appErrorClass;
            }
          oldCatNamesPtr = MemPtrNew(sizeof(Byte) * numEntries[recIdx - 1] * TSMaxCatEntryLen);
          if (!oldCatNamesPtr)
            {
              /* Appear to be out of memory! */
              return appErrorClass;
            }
          /* Open record and move existing entries up */
          voidPtr = MemHandleLock(voidHand);
          MemMove(oldCatNamesPtr,
                  voidPtr + (sizeof(Byte) * 22) /* start past 22 entry transTable */ ,
                  sizeof(Byte) * numEntries[recIdx - 1] * TSMaxCatEntryLen);
          DmWrite(voidPtr,
                  sizeof(Byte) * TSMaxCatEntries,
                  oldCatNamesPtr,
                  sizeof(Byte) * numEntries[recIdx - 1] * TSMaxCatEntryLen);
          MemPtrFree(oldCatNamesPtr);
          /* Fix up the translation table entries */
          for (idx = 1; idx < 22; idx ++)
            {
              /* Covert any references to old max idx (22) to new max idx (102) */
              if (*(((Byte*)voidPtr) + idx) == 22)
                {
                  DmSet(voidPtr, sizeof(Byte) * idx, sizeof(Byte), TSMaxCatEntries);
                }
            }
          DmSet(voidPtr,
                sizeof(Byte) * (21 /* Want to overwrite existing 'Edit...' index */),
                sizeof(Byte) * (TSMaxCatEntries - 21),
                TSMaxCatEntries);
          /* Fixup the 'Edit...' transtable entry that always appears at the end of the transtable */
          DmSet(voidPtr, sizeof(Byte) * (TSMaxCatEntries - 1), 1, numEntries[recIdx - 1] - 1);
          MemHandleUnlock(voidHand);
        }
      /* Database has been upgraded to 1.5 format */
      fromDBVersion = 0x0105 /* Timesheet database version 1.5. */ ;
      DmSetDatabaseInfo(cardNo, dbID, NULL, NULL, &fromDBVersion, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
    }
  else
    {
      /* Don't know how to upgrade this database version! */
      return appErrorClass;
    }
  /* Upgraded database ok */
  return 0;
}

/******************************************************************
 * Check the version of an existing Timesheet database is compatible
 * with the database version the application expects. If not, attempt
 * to upgrade the database, or return an error if we can't.
 ******************************************************************/
Err
VersionCheckDatabase(void)
{
  Err error;
  LocalID dbID;
  UInt dbVer;
  UInt cardNo;

  /* get the database version number */
  DmOpenDatabaseInfo(TSDatabase, &dbID, NULL, NULL, &cardNo, NULL);
  DmDatabaseInfo(cardNo, dbID, NULL, NULL, &dbVer, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
  /* check the database version is compatible with the database version Timesheet expects */
  if (dbVer != TSDBVersion)
    {
      if (FrmAlert(alertID_tsDBConversion) == 1)
      {
        /* User cancelled dialog, exit Timesheet without upgrading database */
        return appErrorClass;
      }
      /* existing database differs from the version application requires */
      error = UpgradeDatabase(cardNo, dbID, dbVer);
      if (error)
        {
          FrmAlert(alertID_tsDBVerIncompat);
          return error;
        }
    }
  /* database looks good */
  return 0;
}

/*** End of file ***/