/******************************************************************
 * PalmOS includes.
 ******************************************************************/
#include <Common.h>
#include <System/SysAll.h>
#include <System/SysEvtMgr.h>
#include <UI/UIAll.h>

/******************************************************************
 * Application includes.
 ******************************************************************/
#include "callback.h"
#include "resource.h"

#include "Timesheet.h"
#include "Globals.h"
#include "Category.h"
#include "DetailForm.h"
#include "MainForm.h"
#include "UIUtil.h"

/******************************************************************
 * Add the timed time to the detail form entry time
 ******************************************************************/
void DetailFormAddTime(long deltaSecs)
{
  Byte entryHours;
  Byte entryMins;
  Long deltaHours;
  Long deltaMins;
  Long deltaFiveMins;

  /* Convert the seconds time to hours and minutes */
  deltaHours = deltaSecs / 3600;
  deltaMins = ( deltaSecs - (deltaHours * 3600) ) / 60;
  deltaFiveMins = deltaMins / 5;

  /* Check we have sufficient time to add */
  if (deltaHours == 0 && deltaFiveMins == 0)
    {
      /* Not enough time has elapsed to add, alert the user */
      FrmAlert(alertID_tsTimerTooShort);
    }
  else
    {
      /* Sufficient time has elapsed to add */

      /* Extract current hours / minutes from details form */
      DetailFormGetDuration(FrmGetActiveForm(), &entryHours, &entryMins);

      /* Add delta hours to entry hours */
      entryMins += deltaFiveMins;
      if (entryMins >= 12)
        {
          entryMins -= 12;
          entryHours ++;
        }

      /* Check if we've exceeded time limit for single entry */
      if (entryHours + deltaHours > 15)
      {
        /* Exceeded maximum representable time in current database format, alert user */
        FrmAlert(alertID_tsTimerTooLong);
        entryHours = 15;
        entryMins = 11;
      }
      else
      {
        entryHours += deltaHours;
      }

      /* Save current hours / minutes back to details form */
      DetailFormSetDuration(FrmGetActiveForm(), entryHours, entryMins);
    }
}

/******************************************************************
 * Close the detail form, releasing any dynamic memory used.
 ******************************************************************/
void
DetailFormClose(FormPtr formPtr)
{
  UInt id;
  ListPtr listPtr;
  FieldPtr fieldPtr;
  VoidHand hand;

  /* free dynamically allocated field memory */
  fieldPtr = FrmGetObjectPtr(formPtr, FrmGetObjectIndex(formPtr, fieldID_tsDesc));
  hand = (VoidHand) FldGetTextHandle(fieldPtr);
  if (hand)
    {
      /* set handle to null so form closing doesn't cause problems */
      FldSetTextHandle(fieldPtr, 0);
      /* release the dynamically allocated memory */
      MemHandleFree(hand);
    }
  /* free the memory used by popup list text ptr arrays */
  for (id = listID_tsClient; id <= listID_tsTask; id++)
    {
      listPtr = FrmGetObjectPtr(formPtr, FrmGetObjectIndex(formPtr, id));
      MemPtrFree(listPtr->itemsText);
      listPtr->itemsText = NULL;
    }
}

/******************************************************************
 * Delete the currently selected entry record from the timesheet
 * database. May also delete the current day (if the last entry for
 * the day is deleted). Returns the number of entries LEFT in the
 * day after the deletion (0 if day has been deleted).
 ******************************************************************/
UInt
DetailFormDeleteEntry(UInt entryRecIdx)
{
  Byte numEntries;
  Byte entryNum;
  UInt idx;
  VoidHand recHand;
  TSDayRecType *dayPtr;
  TSEntryRecType *entryPtr;

  /* delete the current entry from the database */
  DmRemoveRecord(TSDatabase, entryRecIdx);
  /* deleted an entry from the present day */
  recHand = DmQueryRecord(TSDatabase, TSPresentDayRecIdx);
  dayPtr = MemHandleLock(recHand);
  numEntries = (dayPtr->numEntries) - 1;
  /* any entries left for present day? */
  if (numEntries == 0)
    {
      /* no, delete day record... */
      MemHandleUnlock(recHand);
      DmRemoveRecord(TSDatabase, TSPresentDayRecIdx);
      /* and update globals accordingly */
      TSPresentDayRecExists = false;
    }
  else
    {
      /* yes, leave day record intact (but save the new # entries) */
      DmWrite(dayPtr, 0, &numEntries, sizeof(Byte));
      MemHandleUnlock(recHand);
      /* fix all the entries after the deleted entry by reducing their entryNum by one */
      for (idx = entryRecIdx; idx <= TSPresentDayRecIdx + numEntries; idx++)
        {
          recHand = DmGetRecord(TSDatabase, idx);
          entryPtr = MemHandleLock(recHand);
          entryNum = entryPtr->entryNum;
          entryNum--;
          DmWrite(entryPtr, (VoidPtr) (&(entryPtr->entryNum)) - (VoidPtr) entryPtr, &entryNum, sizeof(Byte));
          MemHandleUnlock(recHand);
          DmReleaseRecord(TSDatabase, idx, true);
        }
    }
  /* Check if we've deleted the timer entry record (if any) */
  if (TSEntryTimer.recIdx > 0)
  {
    if (TSEntryTimer.recIdx == entryRecIdx)
    {
      /* Stop the timer as the related record has gone */
      TSEntryTimer.recIdx = 0;
      TSEntryTimer.secs = 0;
    }
    else if (TSEntryTimer.recIdx > entryRecIdx)
    {
      /* Adjust the timer record down to account for deleted record */
      if (numEntries == 0)
      {
        /* Deleted entry AND day record that contained entry */
        TSEntryTimer.recIdx -= 2;
      }
      else
      {
        /* Just deleted a single entry */
        TSEntryTimer.recIdx -= 1;
      }
    }
  }
  return numEntries;
}

/******************************************************************
 * Event handler for the Details form.
 ******************************************************************/
Boolean
DetailFormEventHandler(EventPtr event)
{
  ControlPtr controlPtr;
  FormPtr formPtr;
  RectangleType rect;
  Boolean handled = false;

  if (event->eType == fldChangedEvent)
    {
      /* desc field has changed, update scroll buttons */
      DetailFormScrollBarUpdate(FrmGetActiveForm());
      handled = true;
    }
  else if (event->eType == sclRepeatEvent)
    {
      /* scroll the field */
      DetailFormScroll(FrmGetActiveForm(), event->data.sclRepeat.newValue - event->data.sclRepeat.value);
    }
  else if (event->eType == ctlSelectEvent)
    {
      /* a control button was pressed and released */
      if (event->data.ctlSelect.controlID == buttonID_tsOk)
        {
          /* add a new entry or modify an existing one */
          if (TSEntryTimer.secs > 0 && TSEntryTimer.recIdx == TSEditEntryIdx)
            {
              /* We have a timer running, does the user want to stop it or keep timing? */
              if ((TSAppPrefs.prefFlags & TSPrefAlwaysTimeFlag)
                  || FrmAlert(alertID_tsKeepTiming) == 0)
              {
                /* Do nothing, keep timing */
              }
              else
              {
                /* Stop entry timer and update form with time elapsed prior to entry update */
                DetailFormAddTime((TimGetSeconds() - TSEntryTimer.secs) + 1);
                TSEntryTimer.secs = 0;
                TSEntryTimer.recIdx = 0;
              }
            }
          if (TSEditEntryIdx == 0)
            {
              /* add a new entry, updating timing with the new entry index */
              if (TSEntryTimer.secs > 0 && TSEntryTimer.recIdx == TSEditEntryIdx)
              {
                TSEntryTimer.recIdx = DetailFormNewEntry();
              }
              else
              {
                DetailFormNewEntry();
              }
            }
          else
            {
              /* modify the existing, selected entry */
              DetailFormToEntryRec(TSEditEntryIdx);
            }
          /* show the main form again */
          DetailFormClose(FrmGetActiveForm());
          FrmReturnToForm(formID_tsMain);
          MainFormRefresh();
          handled = true;
        }
      else if (event->data.ctlSelect.controlID == buttonID_tsCancel)
        {
          if (TSEntryTimer.secs > 0 && TSEntryTimer.recIdx == TSEditEntryIdx)
          {
            /* Stop timing the current entry */
            TSEntryTimer.secs = 0;
            TSEntryTimer.recIdx = 0;
          }
          /* cancel details form and show the main form again */
          DetailFormClose(FrmGetActiveForm());
          FrmReturnToForm(formID_tsMain);
          MainFormRefresh();
          handled = true;
        }
      else if (event->data.ctlSelect.controlID == buttonID_tsDelete)
        {
          /* delete the current entry record (and possibly the current day) */
          /* first confirm deletion */
          if (FrmAlert(alertID_tsConfirmDel) == 0)
            {
              if (TSEntryTimer.secs > 0 && TSEntryTimer.recIdx == TSEditEntryIdx)
              {
               /* Stop timing the current entry */
               TSEntryTimer.secs = 0;
               TSEntryTimer.recIdx = 0;
              }
              /* delete the current entry */
              DetailFormClose(FrmGetActiveForm());
              FrmReturnToForm(formID_tsMain);
              DetailFormDeleteEntry(TSEditEntryIdx);
              /* adjust the table top entry as we've just deleted record */
              if (TSTableTopEntry > 0)
                {
                  TSTableTopEntry--;
                }
              MainFormRefresh();
            }                   /* else do nothing, deletion wasn't confirmed */
          handled = true;
        }
      else if (event->data.ctlSelect.controlID == buttonID_tsStartStop)
        {
          if (TSEntryTimer.secs == 0)
            {
              /* Start entry timer */
              TSEntryTimer.secs = TimGetSeconds();
              TSEntryTimer.recIdx = TSEditEntryIdx;
            }
          else if (TSEntryTimer.recIdx == TSEditEntryIdx)
            {
              /* Stop entry timer */
              formPtr = FrmGetActiveForm();
              controlPtr = FrmGetObjectPtr(formPtr, FrmGetObjectIndex(formPtr, bitmapID_tsStopWatch));
              DetailFormDrawStopWatch(formPtr, controlPtr);
              DetailFormAddTime((TimGetSeconds() - TSEntryTimer.secs) + 1);
              TSEntryTimer.secs = 0;
              TSEntryTimer.recIdx = 0;
            }
          else
            {
              /* There's currently a timer running on another event */
              FrmAlert(alertID_tsTimerOnOtherEntry);
            }
          handled = true;
        }
    }
  else if (event->eType == popSelectEvent)
    {
      if (event->data.popSelect.listID == listID_tsClient)
        {
          /* entry in client list has been selected */
          if (event->data.popSelect.selection == (TSAppPrefs.numCatEntries[0] - 1))
            {
              /* last entry (the 'Edit...' entry) has been selected */
              formPtr = FrmGetActiveForm();
              controlPtr = FrmGetObjectPtr(formPtr, FrmGetObjectIndex(formPtr, listID_tsClient));
              LstSetSelection((ListPtr) controlPtr, event->data.popSelect.priorSelection);
              /* show the edit form */
              TSEditCatRecIdx = 1;
              FrmPopupForm(formID_tsEditCat);
            }
        }
      else if (event->data.popSelect.listID == listID_tsProject)
        {
          /* entry in client list has been selected */
          if (event->data.popSelect.selection == (TSAppPrefs.numCatEntries[1] - 1))
            {
              /* last entry (the 'Edit...' entry) has been selected */
              formPtr = FrmGetActiveForm();
              controlPtr = FrmGetObjectPtr(formPtr, FrmGetObjectIndex(formPtr, listID_tsProject));
              LstSetSelection((ListPtr) controlPtr, event->data.popSelect.priorSelection);
              /* show the edit form */
              TSEditCatRecIdx = 2;
              FrmPopupForm(formID_tsEditCat);
              handled = true;
            }
        }
      else if (event->data.popSelect.listID == listID_tsTask)
        {
          /* entry in client list has been selected */
          if (event->data.popSelect.selection == (TSAppPrefs.numCatEntries[2] - 1))
            {
              /* last entry (the 'Edit...' entry) has been selected */
              formPtr = FrmGetActiveForm();
              controlPtr = FrmGetObjectPtr(formPtr, FrmGetObjectIndex(formPtr, listID_tsTask));
              LstSetSelection((ListPtr) controlPtr, event->data.popSelect.priorSelection);
              /* show the edit form */
              TSEditCatRecIdx = 3;
              FrmPopupForm(formID_tsEditCat);
              handled = true;
            }
        }
    }
  else if (event->eType == menuEvent)
    {
      /* a menu item has been selected (only menu options presently are Undo/Cut/Copy/Paste) */
      formPtr = FrmGetActiveForm();
      controlPtr = FrmGetObjectPtr(formPtr, FrmGetObjectIndex(formPtr, fieldID_tsDesc));
      if (event->data.menu.itemID == menuitemID_tsUndo)
        {
          FldUndo((FieldPtr) controlPtr);
        }
      else if (event->data.menu.itemID == menuitemID_tsCut)
        {
          FldCut((FieldPtr) controlPtr);
        }
      else if (event->data.menu.itemID == menuitemID_tsCopy)
        {
          FldCopy((FieldPtr) controlPtr);
        }
      else if (event->data.menu.itemID == menuitemID_tsPaste)
        {
          FldPaste((FieldPtr) controlPtr);
        }
      handled = true;
    }
  else if (event->eType == nilEvent)
    {
      /* If we're timing, toggle the stopwatch on/off once per second */
      if (TSEntryTimer.secs > 0 && TSEntryTimer.recIdx == TSEditEntryIdx)
        {
          formPtr = FrmGetActiveForm();
          controlPtr = FrmGetObjectPtr(formPtr, FrmGetObjectIndex(formPtr, bitmapID_tsStopWatch));
          if (TimGetSeconds() & 0x01)
            {
              DetailFormDrawStopWatch(formPtr, controlPtr);
            }
          else
            {
              /* Erase the bitmap */
              rect.topLeft.x = controlPtr->bounds.topLeft.x;
              rect.topLeft.y = controlPtr->bounds.topLeft.y;
              rect.extent.x = 7;
              rect.extent.y = 8;
              WinEraseRectangle(&rect, 0);
            }
        }
    }
  else if (event->eType == frmOpenEvent)
    {
      /* the Detail form has been opened */
      formPtr = FrmGetActiveForm();
      DetailFormInit(formPtr);
      FrmDrawForm(formPtr);
      FormUpdateDay(&TSPresentDayDate);
      FrmSetFocus(formPtr, FrmGetObjectIndex(formPtr, fieldID_tsDesc));
      handled = true;
    }

  return handled;
}

/******************************************************************
 * Utility function to save a little repeated code. Draw the stopwatch bitmap
 ******************************************************************/
void DetailFormDrawStopWatch(FormPtr formptr, ControlPtr controlPtr)
{
  BitmapPtr bitmapPtr;
  VoidHand hand;

  /* Draw the bitmap */
  hand = DmGet1Resource('Tbmp', bitmapID_tsStopWatch);
  bitmapPtr = MemHandleLock(hand);
  WinDrawBitmap(bitmapPtr, controlPtr->bounds.topLeft.x, controlPtr->bounds.topLeft.y);
  MemHandleUnlock(hand);
  DmReleaseResource(hand);
}

/******************************************************************
 * Initialise the details form, which is about to be shown.
 ******************************************************************/
void
DetailFormInit(FormPtr formPtr)
{
  UInt recAttr;
  Byte clientIdx;
  Byte projectIdx;
  Byte taskIdx;
  Byte hourIdx;
  Byte minuteIdx;
  TSEntryRecType *entryPtr;
  VoidHand entryHand;
  VoidHand hand;
  VoidHand catHand;
  Ptr ptr;
  TSCatRecType *catPtr;
  CharPtr str;
  ControlPtr controlPtr;
  ListPtr clientListPtr;
  ListPtr projectListPtr;
  ListPtr taskListPtr;
  FieldAttrType attr;

  /* set the client list for the form */
  clientListPtr = DetailFormInitPopupList(listID_tsClient, listpopupID_tsClient, 1, TSAppPrefs.numCatEntries[0]);
  /* set the project list for the form */
  projectListPtr = DetailFormInitPopupList(listID_tsProject, listpopupID_tsProject, 2, TSAppPrefs.numCatEntries[1]);
  /* set the task list for the form */
  taskListPtr = DetailFormInitPopupList(listID_tsTask, listpopupID_tsTask, 3, TSAppPrefs.numCatEntries[2]);
  /* editing new or existing entry? */
  if (TSEditEntryIdx == 0)
    {
      /* we are editing a potentially *new* entry in the present day */
      /* make the delete button invisible */
      controlPtr = FrmGetObjectPtr(formPtr, FrmGetObjectIndex(formPtr, buttonID_tsDelete));
      CtlSetUsable(controlPtr, false);
      /* set the client/project/task indexes depending on the Auto Categories preference */
      if (TSAppPrefs.prefFlags & TSPrefAutoCatsFlag)
        {
          clientIdx = TSAppPrefs.newCatEntryIdx[0];
          projectIdx = TSAppPrefs.newCatEntryIdx[1];
          taskIdx = TSAppPrefs.newCatEntryIdx[2];
        }
      else
        {
          clientIdx = 0;
          projectIdx = 0;
          taskIdx = 0;
        }
      /* set the hours and minutes depending on the Auto Durations preference */
      if (TSAppPrefs.prefFlags & TSPrefAutoDurFlag)
        {
          hourIdx = (TSAppPrefs.newCatHours) & 0x0F;    /* hours occupy low 4 bits */
          minuteIdx = (TSAppPrefs.newCatHours) >> 4;    /* minutes occupy high 4 bits */
        }
      else
        {
          hourIdx = 0;
          minuteIdx = 0;
        }
      /* set the checkbox on/off depending on the global app prefs */
      CtlSetValue(FrmGetObjectPtr(formPtr, FrmGetObjectIndex(formPtr, checkboxID_tsCharge)),
                  (TSAppPrefs.prefFlags & TSPrefDefaultChargeFlag));
      /* make the field blank */
      controlPtr = FrmGetObjectPtr(formPtr, FrmGetObjectIndex(formPtr, fieldID_tsDesc));
      hand = MemHandleNew(sizeof(Char));
      ptr = MemHandleLock(hand);
      MemSet(ptr, 1, 0);        /* make sure the field is blank */
      MemHandleUnlock(hand);
      FldSetTextHandle((FieldPtr) controlPtr, (Handle) hand);
    }
  else
    {
      /* we are editing an exiting entry in the present day */
      entryHand = DmQueryRecord(TSDatabase, TSEditEntryIdx);
      entryPtr = MemHandleLock(entryHand);
      /* set the client list index */
      clientIdx = entryPtr->clientIdx;
      /* set the project list index */
      projectIdx = entryPtr->projectIdx;
      /* set the task list index */
      taskIdx = entryPtr->taskIdx;
      /* set the hours and minutes */
      hourIdx = (entryPtr->hours) & 0x0F;       /* hours occupy low 4 bits */
      minuteIdx = (entryPtr->hours) >> 4;       /* minutes occupy high 4 bits */
      /* set the checkbox on/off depending on entry chargeable bit */
      DmRecordInfo(TSDatabase, TSEditEntryIdx, &recAttr, NULL, NULL);
      CtlSetValue(FrmGetObjectPtr(formPtr, FrmGetObjectIndex(formPtr, checkboxID_tsCharge)), (recAttr & TSEntryChargedAttr) >> 3);
      /* set the desc field using a dynamic handle */
      controlPtr = FrmGetObjectPtr(formPtr, FrmGetObjectIndex(formPtr, fieldID_tsDesc));
      str = (CharPtr) (((VoidPtr) entryPtr) + sizeof(TSEntryRecType));  /* desc starts after structure in memory */
      hand = MemHandleNew(sizeof(Char) * (StrLen(str) + 1));
      /* copy into dynamic mem the existing string */
      ptr = MemHandleLock(hand);
      StrCopy((CharPtr) ptr, str);
      MemHandleUnlock(hand);
      FldSetTextHandle((FieldPtr) controlPtr, (Handle) hand);
      /* finished setting up the form for existing entry */
      MemHandleUnlock(entryHand);
    }
  /* have the field send event to maintain the scroll bar. */
  FldGetAttributes((FieldPtr) controlPtr, &attr);
  attr.hasScrollBar = true;
  FldSetAttributes((FieldPtr) controlPtr, &attr);
  /* set the client list */
  catHand = DmQueryRecord(TSDatabase, 1);
  catPtr = MemHandleLock(catHand);
  clientIdx = catPtr->transTable[clientIdx];
  MemHandleUnlock(catHand);
  LstSetSelection(clientListPtr, clientIdx);
  controlPtr = FrmGetObjectPtr(formPtr, FrmGetObjectIndex(formPtr, listpopupID_tsClient));
  CtlSetLabel(controlPtr, LstGetSelectionText(clientListPtr, clientIdx));
  /* set the project list */
  catHand = DmQueryRecord(TSDatabase, 2);
  catPtr = MemHandleLock(catHand);
  projectIdx = catPtr->transTable[projectIdx];
  MemHandleUnlock(catHand);
  LstSetSelection(projectListPtr, projectIdx);
  controlPtr = FrmGetObjectPtr(formPtr, FrmGetObjectIndex(formPtr, listpopupID_tsProject));
  CtlSetLabel(controlPtr, LstGetSelectionText(projectListPtr, projectIdx));
  /* set the task list */
  catHand = DmQueryRecord(TSDatabase, 3);
  catPtr = MemHandleLock(catHand);
  taskIdx = catPtr->transTable[taskIdx];
  MemHandleUnlock(catHand);
  LstSetSelection(taskListPtr, taskIdx);
  controlPtr = FrmGetObjectPtr(formPtr, FrmGetObjectIndex(formPtr, listpopupID_tsTask));
  CtlSetLabel(controlPtr, LstGetSelectionText(taskListPtr, taskIdx));
  /* set the hours and minutes */
  DetailFormSetDuration(formPtr, hourIdx, minuteIdx);
  /* initialise the scroll bar */
  DetailFormScrollBarUpdate(formPtr);
}

/******************************************************************
 * A new entry has been added at the specified category index.
 ******************************************************************/
void
DetailFormPopupListAdd(UInt listID, UInt popupID, UInt catRecIdx, UInt numCatEntries, UInt addedCatIdx)
{
  Word selected;
  FormPtr formPtr;
  ListPtr listPtr;

  /* get the list's current selection */
  formPtr = FrmGetFormPtr(formID_tsDetail);
  listPtr = FrmGetObjectPtr(formPtr, FrmGetObjectIndex(formPtr, listID));
  selected = LstGetSelection(listPtr);
  /* re-fill the list (including the added entry) */
  DetailFormFillPopupList(listID, popupID, catRecIdx, numCatEntries, true);
  /* reselect the appropriate selection */
  if (selected >= addedCatIdx)
    {
      selected += 1;
    }
  LstSetSelection(listPtr, selected);
}

/******************************************************************
 * An existing entry has been removed at the specified category index.
 ******************************************************************/
void
DetailFormPopupListDelete(UInt listID, UInt popupID, UInt catRecIdx, UInt numCatEntries, UInt deletedCatIdx)
{
  Word selected;
  FormPtr formPtr;
  ListPtr listPtr;

  /* get the list's current selection */
  formPtr = FrmGetFormPtr(formID_tsDetail);
  listPtr = FrmGetObjectPtr(formPtr, FrmGetObjectIndex(formPtr, listID));
  selected = LstGetSelection(listPtr);
  /* re-fill the list (including the added entry) */
  DetailFormFillPopupList(listID, popupID, catRecIdx, numCatEntries, true);
  /* reselect the appropriate selection */
  if (selected == deletedCatIdx)
    {
      selected = 0;
    }
  else if (selected > deletedCatIdx)
    {
      selected -= 1;
    }
  LstSetSelection(listPtr, selected);
}

/******************************************************************
 * An existing entry has been renamed at the specified category index.
 ******************************************************************/
void
DetailFormPopupListRename(UInt listID, UInt popupID, UInt catRecIdx, UInt numCatEntries,
                          UInt renamedCatIdx, UInt renameToCatIdx)
{
  Word selected;
  FormPtr formPtr;
  ListPtr listPtr;

  /* get the list's current selection */
  formPtr = FrmGetFormPtr(formID_tsDetail);
  listPtr = FrmGetObjectPtr(formPtr, FrmGetObjectIndex(formPtr, listID));
  selected = LstGetSelection(listPtr);
  /* re-fill the list (including the added entry) */
  DetailFormFillPopupList(listID, popupID, catRecIdx, numCatEntries, true);
  /* reselect the appropriate selection */
  if (renamedCatIdx == renameToCatIdx)
    {
      /* do nothing, renamed category hasn't changed place */
      return;
    }
  else if (selected == renamedCatIdx)
    {
      /* change selection to new position of renamed cat */
      selected = renameToCatIdx;
    }
  LstSetSelection(listPtr, selected);
}

/******************************************************************
 * Fill the supplied popup list with entries from the specified category.
 * If freeExisting is true, the memory used by existing list entries if released first.
 ******************************************************************/
ListPtr
DetailFormFillPopupList(UInt listID, UInt popupID, UInt catRecIdx, UInt numCatEntries,
                        Boolean freeExisting)
{
  VoidHand catHand;
  TSCatRecType *catPtr;
  CharPtr *catStrPtr;
  ListPtr listPtr;
  FormPtr formPtr;

  /* get the list control */
  formPtr = FrmGetFormPtr(formID_tsDetail);
  listPtr = FrmGetObjectPtr(formPtr, FrmGetObjectIndex(formPtr, listID));
  /* open the category record */
  catHand = DmGetRecord(TSDatabase, catRecIdx);
  catPtr = MemHandleLock(catHand);
  /* free memory for old pointer array if it exists */
  if (freeExisting)
    {
      MemPtrFree(listPtr->itemsText);
      listPtr->itemsText = NULL;
    }
  /* rebuild the pointer array in case OS has moved records around in memory */
  catStrPtr = CatBuildPtrArray(catPtr, numCatEntries);
  /* fill it with the supplied entries */
  LstSetListChoices(listPtr, catStrPtr, numCatEntries);
  /* set the list length depending on the number of entries so we don't show funny looking half-empty lists */
  if (numCatEntries < 10)
    {
      LstSetHeight(listPtr, numCatEntries);
    }
  else
    {
      LstSetHeight(listPtr, 10);
    }
  /* release the category record */
  MemHandleUnlock(catHand);
  DmReleaseRecord(TSDatabase, catRecIdx, true);
  return listPtr;
}

/******************************************************************
 * Set up the details control for the supplied list.
 ******************************************************************/
ListPtr
DetailFormInitPopupList(UInt listID, UInt popupID, UInt catRecIdx, UInt numCatEntries)
{
  ListPtr listPtr;
  ControlPtr popupPtr;
  FormPtr formPtr;

  /* fill the list */
  listPtr = DetailFormFillPopupList(listID, popupID, catRecIdx, numCatEntries, false);
  /* make the list and the matching popup display the first list item */
  formPtr = FrmGetFormPtr(formID_tsDetail);
  LstSetTopItem(listPtr, 0);
  popupPtr = FrmGetObjectPtr(formPtr, FrmGetObjectIndex(formPtr, popupID));
  CtlSetLabel(popupPtr, LstGetSelectionText(listPtr, 0));
  return listPtr;
}

/******************************************************************
 * User has entered new entry details and clicked ok, time to add a
 * new entry to present day.
 ******************************************************************/
UInt
DetailFormNewEntry(void)
{
  Byte numEntries;
  UInt entryIdx;
  UInt attr;
  Err error;
  VoidHand dayHand;
  VoidHand entryHand;
  TSDayRecType day;
  TSDayRecType *dayPtr;
  TSEntryRecType *entryPtr;

  /* is there an existing record for this day? */
  if (!TSPresentDayRecExists)
    {
      /* no, so create one before creating the initial entry for the day */
      dayHand = DmNewRecord(TSDatabase, &TSPresentDayRecIdx, sizeof(TSDayRecType));
      /* fill the day record with appropriate details */
      dayPtr = MemHandleLock(dayHand);
      day.dayDate.day = TSPresentDayDate.day;
      day.dayDate.month = TSPresentDayDate.month;
      day.dayDate.year = TSPresentDayDate.year;
      day.numEntries = 1;       /* just about to add first record */
      numEntries = 1;
      DmWrite(dayPtr, 0, &day, sizeof(TSDayRecType));
      MemHandleUnlock(dayHand);
      DmReleaseRecord(TSDatabase, TSPresentDayRecIdx, true);
      /* mark the record as a day record */
      attr = TSDayRecAttr;
      error = DmSetRecordInfo(TSDatabase, TSPresentDayRecIdx, &attr, NULL);
      /* created new day record */
      TSPresentDayRecExists = true;
      /* set the next new entry index to the first record after this day */
      entryIdx = TSPresentDayRecIdx + 1;
      /* Check if we need to update the entry timer record index */
      if (TSEntryTimer.recIdx >= TSPresentDayRecIdx)
      {
        TSEntryTimer.recIdx += 1;
      }
    }
  else
    {
      /* yes, update the present day and find the point to insert the next entry record for the day */
      dayHand = DmGetRecord(TSDatabase, TSPresentDayRecIdx);
      dayPtr = MemHandleLock(dayHand);
      numEntries = dayPtr->numEntries;
      numEntries += 1;
      DmWrite(dayPtr, 0, &numEntries, sizeof(Byte));
      entryIdx = TSPresentDayRecIdx + numEntries;
      MemHandleUnlock(dayHand);
      DmReleaseRecord(TSDatabase, TSPresentDayRecIdx, true);
    }
  /* create a new database record */
  DmNewRecord(TSDatabase, &entryIdx, sizeof(TSEntryRecType) + 1);
  DmReleaseRecord(TSDatabase, entryIdx, true);
  /* mark the record as a day record */
  attr = TSEntryRecAttr;
  error = DmSetRecordInfo(TSDatabase, entryIdx, &attr, NULL);
  /* fill the entry record with details from the database */
  DetailFormToEntryRec(entryIdx);
  /* fill in the new record's day entry index */
  entryHand = DmGetRecord(TSDatabase, entryIdx);
  entryPtr = MemHandleLock(entryHand);
  DmWrite(entryPtr, (VoidPtr) & (entryPtr->entryNum) - (VoidPtr) entryPtr, &numEntries, sizeof(Byte));
  MemHandleUnlock(entryHand);
  DmReleaseRecord(TSDatabase, entryIdx, true);
  /* make sure the new entry will be visible in the main table */
  if (TSTableTopEntry + TSMainFormTableRows < numEntries)
    {
      /* need to scroll table up to show added entry */
      TSTableTopEntry = numEntries - TSMainFormTableRows;
    }
  /* Check if we need to update the entry timer record index */
  if (TSEntryTimer.recIdx >= entryIdx)
  {
    TSEntryTimer.recIdx += 1;
  }
  /* Return the new entry's database index */
  return entryIdx;
}

/******************************************************************
 * Handle scrolling the desc field on the detail form, driven by
 * scrollbar events.
 ******************************************************************/
void
DetailFormScroll(FormPtr formPtr, Int lines)
{
  Word blankLines;
  Short min;
  Short max;
  Short value;
  Short pageSize;
  FieldPtr fld;
  ScrollBarPtr bar;

  fld = FrmGetObjectPtr(formPtr, FrmGetObjectIndex(formPtr, fieldID_tsDesc));

  if (lines < 0)
    {
      blankLines = FldGetNumberOfBlankLines(fld);
      FldScrollField(fld, -lines, up);

      /* If there were blank lines visible at the end of the field
       * then we need to update the scroll bar. */
      if (blankLines)
        {
          /* Update the scroll bar. */
          bar = FrmGetObjectPtr(formPtr, FrmGetObjectIndex(formPtr, scrollbarID_tsDesc));
          SclGetScrollBar(bar, &value, &min, &max, &pageSize);
          if (blankLines > -lines)
            max += lines;
          else
            max -= blankLines;
          SclSetScrollBar(bar, value, min, max, pageSize);
        }
    }
  else if (lines > 0)
    FldScrollField(fld, lines, down);
}

/******************************************************************
 * Update the form's scroll bar to reflect changes in the desc field.
 ******************************************************************/
void
DetailFormScrollBarUpdate(FormPtr formPtr)
{
  Word scrollPos;
  Word textHeight;
  Word fieldHeight;
  Short maxValue;
  FieldPtr fld;
  ScrollBarPtr bar;

  fld = FrmGetObjectPtr(formPtr, FrmGetObjectIndex(formPtr, fieldID_tsDesc));
  bar = FrmGetObjectPtr(formPtr, FrmGetObjectIndex(formPtr, scrollbarID_tsDesc));

  FldGetScrollValues(fld, &scrollPos, &textHeight, &fieldHeight);

  if (textHeight > fieldHeight)
    maxValue = textHeight - fieldHeight;
  else if (scrollPos)
    maxValue = scrollPos;
  else
    maxValue = 0;

  SclSetScrollBar(bar, scrollPos, 0, maxValue, fieldHeight - 1);
}

/******************************************************************
 * Fills the supplied database entry record with fields from the
 * presently visible DetailForm.
 ******************************************************************/
void
DetailFormToEntryRec(UInt entryRecIdx)
{
  UInt recAttr;
  UInt descLen;
  ULong entrySize;
  ULong newEntrySize;
  TSEntryRecType entry;
  FormPtr formPtr;
  ControlPtr controlPtr;
  VoidHand hand;
  VoidPtr ptr;
  Byte hours;
  Byte mins;
  Byte clientSelectIdx;
  Byte projectSelectIdx;
  Byte taskSelectIdx;

  /* build a local entry from the GUI control settings */
  formPtr = FrmGetActiveForm();
  /* save the hour setting */
  DetailFormGetDuration(formPtr, &hours, &mins);
  entry.hours = hours + (mins << 4);
  /* save the selected client (if any) */
  controlPtr = FrmGetObjectPtr(formPtr, FrmGetObjectIndex(formPtr, listID_tsClient));
  /* don't save the actual selection idx, save the translation table entry that contains the selection idx instead */
  hand = DmQueryRecord(TSDatabase, 1);
  ptr = MemHandleLock(hand);
  clientSelectIdx = (Byte) LstGetSelection((ListPtr) controlPtr);
  entry.clientIdx = CatReverseLookup(clientSelectIdx, ptr, TSMaxCatEntries);
  MemHandleUnlock(hand);
  /* save the project setting (if any) */
  controlPtr = FrmGetObjectPtr(formPtr, FrmGetObjectIndex(formPtr, listID_tsProject));
  /* don't save the actual selection idx, save the translation table entry that contains the selection idx instead */
  hand = DmQueryRecord(TSDatabase, 2);
  ptr = MemHandleLock(hand);
  projectSelectIdx = (Byte) LstGetSelection((ListPtr) controlPtr);
  entry.projectIdx = CatReverseLookup(projectSelectIdx, ptr, TSMaxCatEntries);
  MemHandleUnlock(hand);
  /* save the task setting (if any) */
  controlPtr = FrmGetObjectPtr(formPtr, FrmGetObjectIndex(formPtr, listID_tsTask));
  /* don't save the actual selection idx, save the translation table entry that contains the selection idx instead */
  hand = DmQueryRecord(TSDatabase, 3);
  ptr = MemHandleLock(hand);
  taskSelectIdx = (Byte) LstGetSelection((ListPtr) controlPtr);
  entry.taskIdx = CatReverseLookup(taskSelectIdx, ptr, TSMaxCatEntries);
  MemHandleUnlock(hand);
  /* if this is a new entry, save the client/project/task selection index for the auto categories option
   * NOTE: we *always* save this info for new entries, but we don't always use it when we init the new detail form */
  if (TSEditEntryIdx == 0)
    {
      /* new entry, save selected client/project/task selection index (NOT the translated index!) into app prefs */
      TSAppPrefs.newCatEntryIdx[0] = entry.clientIdx;
      TSAppPrefs.newCatEntryIdx[1] = entry.projectIdx;
      TSAppPrefs.newCatEntryIdx[2] = entry.taskIdx;
      /* also save the hour/minutes duration into app prefs */
      TSAppPrefs.newCatHours = entry.hours;
    }
  /* find out how big the entry will be (depends on the amount of descriptive text that has been entered by the user) */
  controlPtr = FrmGetObjectPtr(formPtr, FrmGetObjectIndex(formPtr, fieldID_tsDesc));
  descLen = FldGetTextLength((FieldPtr) controlPtr);
  /* resize the record if required to accomodate the entered text */
  hand = DmGetRecord(TSDatabase, entryRecIdx);
  entrySize = MemHandleSize(hand);
  newEntrySize = sizeof(TSEntryRecType) + ((descLen + 1) * sizeof(Byte));
  if (entrySize != newEntrySize)
    {
      /* resize the record (NOTE: handle may change!) */
      DmReleaseRecord(TSDatabase, entryRecIdx, false);
      hand = DmResizeRecord(TSDatabase, entryRecIdx, newEntrySize);
    }
  /* write the entry */
  ptr = MemHandleLock(hand);
  /* save the current entry num */
  entry.entryNum = ((TSEntryRecType *) ptr)->entryNum;
  DmWrite(ptr, 0, &entry, sizeof(TSEntryRecType));
  /* write the descriptive text (if any) */
  if (descLen > 0)
    {
      /* write some actual text */
      DmWrite(ptr, sizeof(TSEntryRecType), FldGetTextPtr((FieldPtr) controlPtr), (descLen + 1) * sizeof(Byte));
    }
  else
    {
      /* make sure a null string gets in there */
      DmSet(ptr, sizeof(TSEntryRecType), 1, 0);
    }
  /* entry has been updated */
  MemHandleUnlock(hand);
  DmReleaseRecord(TSDatabase, entryRecIdx, true);
  /* set the high bit of the record category (low four bits of attribute) on if entry is chargeable */
  DmRecordInfo(TSDatabase, entryRecIdx, &recAttr, NULL, NULL);
  controlPtr = FrmGetObjectPtr(formPtr, FrmGetObjectIndex(formPtr, checkboxID_tsCharge));
  if (CtlGetValue(controlPtr))
    {
      /* turn bit 4 on (high bit of category) */
      recAttr = TSEntryRecAttr | 0x0008;
    }
  else
    {
      /* turn bit 4 off (high bit of category) */
      recAttr = TSEntryRecAttr & 0xFFF7;
    }
  DmSetRecordInfo(TSDatabase, entryRecIdx, &recAttr, NULL);
}

/******************************************************************
 * Detail form utility function to get duration hours/mins (in five mins intervals)
 ******************************************************************/
void DetailFormGetDuration(FormPtr formPtr, Byte* hoursPtr, Byte* fiveMinsPtr)
{
  *hoursPtr = LstGetSelection(FrmGetObjectPtr(formPtr, FrmGetObjectIndex(formPtr, listID_tsHour)));
  *fiveMinsPtr = LstGetSelection(FrmGetObjectPtr(formPtr, FrmGetObjectIndex(formPtr, listID_tsMin)));
}

/******************************************************************
 * Detail form utility function to set duration hours/mins (in five mins intervals)
 ******************************************************************/
void DetailFormSetDuration(FormPtr formPtr, Byte hours, Byte fiveMins)
{
  LstSetSelection(FrmGetObjectPtr(formPtr, FrmGetObjectIndex(formPtr, listID_tsHour)), hours);
  CtlSetLabel(FrmGetObjectPtr(formPtr, FrmGetObjectIndex(formPtr, listpopupID_tsHour)),
      LstGetSelectionText(FrmGetObjectPtr(formPtr, FrmGetObjectIndex(formPtr, listID_tsHour)),
                          hours));
  LstSetSelection(FrmGetObjectPtr(formPtr, FrmGetObjectIndex(formPtr, listID_tsMin)), fiveMins);
  CtlSetLabel(FrmGetObjectPtr(formPtr, FrmGetObjectIndex(formPtr, listpopupID_tsMin)),
              LstGetSelectionText(FrmGetObjectPtr(formPtr, FrmGetObjectIndex(formPtr, listID_tsMin)),
                          fiveMins));
}

/*** End of file ***/